@@ -1,6 +1,6 @@
-/* 
+/*
  * dhcpcd - DHCP client daemon
- * Copyright (c) 2006-2011 Roy Marples <roy@marples.name>
+ * Copyright (c) 2006-2013 Roy Marples <roy@marples.name>
  * All rights reserved
 
  * Redistribution and use in source and binary forms, with or without
@@ -57,8 +57,9 @@
 
 #include "config.h"
 #include "common.h"
-#include "configure.h"
+#include "dev.h"
 #include "dhcp.h"
+#include "ipv4.h"
 #include "ipv6.h"
 #include "net.h"
 
@@ -106,7 +107,7 @@ if_conf(struct interface *iface)
 	if (fp == NULL)
 		return errno == ENOENT ? 0 : -1;
 	if (fgets(buf, sizeof(buf), fp) != NULL && buf[0] == '0')
-		iface->state->options->options |= DHCPCD_BROADCAST;
+		iface->options->options |= DHCPCD_BROADCAST;
 	fclose(fp);
 	return 0;
 }
@@ -122,11 +123,12 @@ _open_link_socket(struct sockaddr_nl *nl)
 	if (bind(fd, (struct sockaddr *)nl, sizeof(*nl)) == -1)
 		return -1;
 	set_cloexec(fd);
+
 	return fd;
 }
 
 int
-init_sockets(void)
+open_sockets(void)
 {
 	if ((socket_afnet = socket(AF_INET, SOCK_DGRAM, 0)) == -1)
 		return -1;
@@ -142,7 +144,15 @@ open_link_socket(void)
 	struct sockaddr_nl snl;
 
 	memset(&snl, 0, sizeof(snl));
-	snl.nl_groups = RTMGRP_LINK | RTMGRP_IPV4_ROUTE | RTMGRP_IPV4_IFADDR;
+	snl.nl_groups = RTMGRP_LINK;
+
+#ifdef INET
+	snl.nl_groups |= RTMGRP_IPV4_ROUTE | RTMGRP_IPV4_IFADDR;
+#endif
+#ifdef INET6
+	snl.nl_groups |= RTMGRP_IPV6_ROUTE | RTMGRP_IPV6_IFADDR;
+#endif
+
 	return _open_link_socket(&snl);
 }
 
@@ -264,11 +274,10 @@ link_route(struct nlmsghdr *nlm)
 		return 1;
 	rta = (struct rtattr *)(void *)((char *)rtm +NLMSG_ALIGN(sizeof(*rtm)));
 	len = NLMSG_PAYLOAD(nlm, sizeof(*rtm));
-	rt.iface = NULL;
+	memset(&rt, 0, sizeof(rt));
 	rt.dest.s_addr = INADDR_ANY;
 	rt.net.s_addr = INADDR_ANY;
 	rt.gate.s_addr = INADDR_ANY;
-	rt.next = NULL;
 	metric = 0;
 	while (RTA_OK(rta, len)) {
 		switch (rta->rta_type) {
@@ -293,8 +302,10 @@ link_route(struct nlmsghdr *nlm)
 	}
 	if (rt.iface != NULL) {
 		if (metric == rt.iface->metric) {
+#ifdef INET
 			inet_cidrtoaddr(rtm->rtm_dst_len, &rt.net);
-			route_deleted(&rt);
+			ipv4_routedeleted(&rt);
+#endif
 		}
 	}
 	return 1;
@@ -306,9 +317,14 @@ link_addr(struct nlmsghdr *nlm)
 	int len;
 	struct rtattr *rta;
 	struct ifaddrmsg *ifa;
-	struct in_addr addr, net, dest;
 	char ifn[IF_NAMESIZE + 1];
 	struct interface *iface;
+#ifdef INET
+	struct in_addr addr, net, dest;
+#endif
+#ifdef INET6
+	struct in6_addr addr6;
+#endif
 
 	if (nlm->nlmsg_type != RTM_DELADDR && nlm->nlmsg_type != RTM_NEWADDR)
 		return 0;
@@ -318,8 +334,8 @@ link_addr(struct nlmsghdr *nlm)
 		errno = EBADMSG;
 		return -1;
 	}
-	if (nlm->nlmsg_pid == (uint32_t)getpid())
-		return 1;
+//	if (nlm->nlmsg_pid == (uint32_t)getpid())
+//		return 1;
 	ifa = NLMSG_DATA(nlm);
 	if (if_indextoname(ifa->ifa_index, ifn) == NULL)
 		return -1;
@@ -328,35 +344,92 @@ link_addr(struct nlmsghdr *nlm)
 		return 1;
 	rta = (struct rtattr *) IFA_RTA(ifa);
 	len = NLMSG_PAYLOAD(nlm, sizeof(*ifa));
-	addr.s_addr = dest.s_addr = INADDR_ANY;
-	dest.s_addr = INADDR_ANY;
-	inet_cidrtoaddr(ifa->ifa_prefixlen, &net);
-	while (RTA_OK(rta, len)) {
-		switch (rta->rta_type) {
-		case IFA_ADDRESS:
-			if (iface->flags & IFF_POINTOPOINT) {
-				memcpy(&dest.s_addr, RTA_DATA(rta),
-				    sizeof(addr.s_addr));
+	switch (ifa->ifa_family) {
+#ifdef INET
+	case AF_INET:
+		addr.s_addr = dest.s_addr = INADDR_ANY;
+		dest.s_addr = INADDR_ANY;
+		inet_cidrtoaddr(ifa->ifa_prefixlen, &net);
+		while (RTA_OK(rta, len)) {
+			switch (rta->rta_type) {
+			case IFA_ADDRESS:
+				if (iface->flags & IFF_POINTOPOINT) {
+					memcpy(&dest.s_addr, RTA_DATA(rta),
+					       sizeof(addr.s_addr));
+				}
+				break;
+			case IFA_LOCAL:
+				memcpy(&addr.s_addr, RTA_DATA(rta),
+				       sizeof(addr.s_addr));
+				break;
 			}
-			break;
-		case IFA_LOCAL:
-			memcpy(&addr.s_addr, RTA_DATA(rta),
-			    sizeof(addr.s_addr));
-			break;
+			rta = RTA_NEXT(rta, len);
 		}
-		rta = RTA_NEXT(rta, len);
+		ipv4_handleifa(nlm->nlmsg_type, NULL, ifn, &addr, &net, &dest);
+		break;
+#endif
+#ifdef INET6
+	case AF_INET6:
+		memset(&addr6, 0, sizeof(addr6));
+		while (RTA_OK(rta, len)) {
+			switch (rta->rta_type) {
+			case IFA_ADDRESS:
+				memcpy(&addr6.s6_addr, RTA_DATA(rta),
+				       sizeof(addr6.s6_addr));
+				break;
+			}
+			rta = RTA_NEXT(rta, len);
+		}
+		ipv6_handleifa(nlm->nlmsg_type, NULL, ifn,
+		    &addr6, ifa->ifa_flags);
+		break;
+#endif
 	}
-	handle_ifa(nlm->nlmsg_type, ifn, &addr, &net, &dest);
 	return 1;
 }
 
+static short l2addr_len(unsigned short if_type)
+{
+
+	switch (if_type) {
+	case ARPHRD_ETHER: /* FALLTHROUGH */
+	case ARPHRD_IEEE802: /*FALLTHROUGH */
+	case ARPHRD_IEEE80211:
+		return 6;
+	case ARPHRD_IEEE1394:
+		return 8;
+	case ARPHRD_INFINIBAND:
+		return 20;
+	default:
+		return -1;
+	}
+}
+
+static int
+handle_rename(unsigned int ifindex, const char *ifname)
+{
+	struct interface *ifp;
+
+	TAILQ_FOREACH(ifp, ifaces, next) {
+		if (ifp->index == ifindex && strcmp(ifp->name, ifname)) {
+			handle_interface(-1, ifp->name);
+			/* Let dev announce the interface for renaming */
+			if (!dev_listening())
+				handle_interface(1, ifname);
+			return 1;
+		}
+	}
+	return 0;
+}
+
 static int
 link_netlink(struct nlmsghdr *nlm)
 {
 	int len;
-	struct rtattr *rta;
+	struct rtattr *rta, *hwaddr;
 	struct ifinfomsg *ifi;
 	char ifn[IF_NAMESIZE + 1];
+	struct interface *ifp;
 
 	len = link_route(nlm);
 	if (len != 0)
@@ -378,6 +451,8 @@ link_netlink(struct nlmsghdr *nlm)
 	rta = (struct rtattr *)(void *)((char *)ifi +NLMSG_ALIGN(sizeof(*ifi)));
 	len = NLMSG_PAYLOAD(nlm, sizeof(*ifi));
 	*ifn = '\0';
+	hwaddr = NULL;
+
 	while (RTA_OK(rta, len)) {
 		switch (rta->rta_type) {
 		case IFLA_WIRELESS:
@@ -389,6 +464,9 @@ link_netlink(struct nlmsghdr *nlm)
 		case IFLA_IFNAME:
 			strlcpy(ifn, RTA_DATA(rta), sizeof(ifn));
 			break;
+		case IFLA_ADDRESS:
+			hwaddr = rta;
+			break;
 		}
 		rta = RTA_NEXT(rta, len);
 	}
@@ -407,7 +485,28 @@ link_netlink(struct nlmsghdr *nlm)
 		return 1;
 	}
 
-	handle_carrier(ifi->ifi_flags & IFF_RUNNING ? 1 : -1,
+	/* Check for interface name change */
+	if (handle_rename(ifi->ifi_index, ifn))
+		    return 1;
+
+	/* Check for a new interface */
+	ifp = find_interface(ifn);
+	if (ifp == NULL) {
+		/* If are listening to a dev manager, let that announce
+		 * the interface rather than the kernel. */
+		if (dev_listening() < 1)
+			handle_interface(1, ifn);
+		return 1;
+	}
+
+	/* Re-read hardware address and friends */
+	if (!(ifi->ifi_flags & IFF_UP) && hwaddr) {
+		len = l2addr_len(ifi->ifi_type);
+		if (hwaddr->rta_len == RTA_LENGTH(len))
+			handle_hwaddr(ifn, RTA_DATA(hwaddr), len);
+	}
+
+	handle_carrier(ifi->ifi_flags & IFF_RUNNING ? LINK_UP : LINK_DOWN,
 	    ifi->ifi_flags, ifn);
 	return 1;
 }
@@ -492,7 +591,7 @@ add_attr_32(struct nlmsghdr *n, unsigned int maxlen, int type, uint32_t data)
 struct nlma
 {
 	struct nlmsghdr hdr;
-	struct ifaddrmsg ifa; 
+	struct ifaddrmsg ifa;
 	char buffer[64];
 };
 
@@ -503,6 +602,7 @@ struct nlmr
 	char buffer[256];
 };
 
+#ifdef INET
 int
 if_address(const struct interface *iface,
     const struct in_addr *address, const struct in_addr *netmask,
@@ -511,7 +611,9 @@ if_address(const struct interface *iface,
 	struct nlma *nlm;
 	int retval = 0;
 
-	nlm = xzalloc(sizeof(*nlm));
+	nlm = calloc(1, sizeof(*nlm));
+	if (nlm == NULL)
+		return -1;
 	nlm->hdr.nlmsg_len = NLMSG_LENGTH(sizeof(struct ifaddrmsg));
 	nlm->hdr.nlmsg_flags = NLM_F_REQUEST;
 	if (action >= 0) {
@@ -542,8 +644,11 @@ if_route(const struct rt *rt, int action)
 {
 	struct nlmr *nlm;
 	int retval = 0;
+	struct dhcp_state *state;
 
-	nlm = xzalloc(sizeof(*nlm));
+	nlm = calloc(1, sizeof(*nlm));
+	if (nlm == NULL)
+		return -1;
 	nlm->hdr.nlmsg_len = NLMSG_LENGTH(sizeof(struct rtmsg));
 	nlm->hdr.nlmsg_type = RTM_NEWROUTE;
 	if (action == 0)
@@ -556,14 +661,15 @@ if_route(const struct rt *rt, int action)
 	nlm->rt.rtm_family = AF_INET;
 	nlm->rt.rtm_table = RT_TABLE_MAIN;
 
+	state = D_STATE(rt->iface);
 	if (action == -1 || action == -2)
 		nlm->rt.rtm_scope = RT_SCOPE_NOWHERE;
 	else {
 		nlm->hdr.nlmsg_flags |= NLM_F_CREATE | NLM_F_EXCL;
 		/* We only change route metrics for kernel routes */
 		if (rt->dest.s_addr ==
-		    (rt->iface->addr.s_addr & rt->iface->net.s_addr) &&
-		    rt->net.s_addr == rt->iface->net.s_addr)
+		    (state->addr.s_addr & state->net.s_addr) &&
+		    rt->net.s_addr == state->net.s_addr)
 			nlm->rt.rtm_protocol = RTPROT_KERNEL;
 		else
 			nlm->rt.rtm_protocol = RTPROT_BOOT;
@@ -581,7 +687,7 @@ if_route(const struct rt *rt, int action)
 	    &rt->dest.s_addr, sizeof(rt->dest.s_addr));
 	if (nlm->rt.rtm_protocol == RTPROT_KERNEL) {
 		add_attr_l(&nlm->hdr, sizeof(*nlm), RTA_PREFSRC,
-		    &rt->iface->addr.s_addr, sizeof(rt->iface->addr.s_addr));
+		    &state->addr.s_addr, sizeof(state->addr.s_addr));
 	}
 	/* If destination == gateway then don't add the gateway */
 	if (rt->dest.s_addr != rt->gate.s_addr ||
@@ -589,7 +695,8 @@ if_route(const struct rt *rt, int action)
 		add_attr_l(&nlm->hdr, sizeof(*nlm), RTA_GATEWAY,
 		    &rt->gate.s_addr, sizeof(rt->gate.s_addr));
 
-	add_attr_32(&nlm->hdr, sizeof(*nlm), RTA_OIF, rt->iface->index);
+	if (rt->gate.s_addr != htonl(INADDR_LOOPBACK))
+		add_attr_32(&nlm->hdr, sizeof(*nlm), RTA_OIF, rt->iface->index);
 	add_attr_32(&nlm->hdr, sizeof(*nlm), RTA_PRIORITY, rt->metric);
 
 	if (send_netlink(&nlm->hdr) == -1)
@@ -597,15 +704,19 @@ if_route(const struct rt *rt, int action)
 	free(nlm);
 	return retval;
 }
+#endif
 
+#ifdef INET6
 int
-if_address6(const struct interface *ifp, const struct ipv6_addr *ap, int action)
+if_address6(const struct ipv6_addr *ap, int action)
 {
 	struct nlma *nlm;
 	struct ifa_cacheinfo cinfo;
 	int retval = 0;
 
-	nlm = xzalloc(sizeof(*nlm));
+	nlm = calloc(1, sizeof(*nlm));
+	if (nlm == NULL)
+		return -1;
 	nlm->hdr.nlmsg_len = NLMSG_LENGTH(sizeof(struct ifaddrmsg));
 	nlm->hdr.nlmsg_flags = NLM_F_REQUEST;
 	if (action >= 0) {
@@ -613,12 +724,12 @@ if_address6(const struct interface *ifp, const struct ipv6_addr *ap, int action)
 		nlm->hdr.nlmsg_type = RTM_NEWADDR;
 	} else
 		nlm->hdr.nlmsg_type = RTM_DELADDR;
-	nlm->ifa.ifa_index = ifp->index;
+	nlm->ifa.ifa_index = ap->iface->index;
 	nlm->ifa.ifa_family = AF_INET6;
 	nlm->ifa.ifa_prefixlen = ap->prefix_len;
 	/* This creates the aliased interface */
 	add_attr_l(&nlm->hdr, sizeof(*nlm), IFA_LABEL,
-	    ifp->name, strlen(ifp->name) + 1);
+	    ap->iface->name, strlen(ap->iface->name) + 1);
 	add_attr_l(&nlm->hdr, sizeof(*nlm), IFA_LOCAL,
 	    &ap->addr.s6_addr, sizeof(ap->addr.s6_addr));
 
@@ -664,7 +775,9 @@ if_route6(const struct rt6 *rt, int action)
 	struct rtattr *metrics = (void *)metricsbuf;
 	int retval = 0;
 
-	nlm = xzalloc(sizeof(*nlm));
+	nlm = calloc(1, sizeof(*nlm));
+	if (nlm == NULL)
+		return -1;
 	nlm->hdr.nlmsg_len = NLMSG_LENGTH(sizeof(struct rtmsg));
 	nlm->hdr.nlmsg_type = RTM_NEWROUTE;
 	nlm->hdr.nlmsg_flags = NLM_F_REQUEST;
@@ -718,3 +831,13 @@ if_route6(const struct rt6 *rt, int action)
 	errno = ENOTSUP;
 	return -1;
 }
+
+int
+in6_addr_flags(__unused const char *ifname,
+    __unused const struct in6_addr *addr)
+{
+
+	/* How do I get IPv6 address flags on Linux? */
+	return 0;
+}
+#endif
